from unittest import TestCase

from py_clob_client.clob_types import (
    OrderArgs,
    MarketOrderArgs,
    CreateOrderOptions,
    OrderSummary,
)
from py_clob_client.constants import AMOY
from py_clob_client.order_builder.constants import BUY, SELL

from py_clob_client.signer import Signer
from py_clob_client.order_builder.builder import OrderBuilder, ROUNDING_CONFIG
from py_clob_client.order_builder.helpers import decimal_places, round_normal
from py_order_utils.model import POLY_GNOSIS_SAFE, EOA

# publicly known private key
private_key = "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
chain_id = AMOY
signer = Signer(private_key=private_key, chain_id=chain_id)


class TestOrderBuilder(TestCase):
    def test_calculate_market_price_buy(self):
        # empty
        with self.assertRaises(Exception):
            builder = OrderBuilder(signer)
            builder.calculate_market_price([], 100)

        # not enough
        with self.assertRaises(Exception):
            positions = [
                OrderSummary(price="0.5", size="100"),
                OrderSummary(price="0.4", size="100"),
            ]
            builder = OrderBuilder(signer)
            builder.calculate_market_price(positions, 100)

        # OK
        positions = [
            OrderSummary(price="0.5", size="100"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.3", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.3)

        positions = [
            OrderSummary(price="0.5", size="100"),
            OrderSummary(price="0.4", size="200"),
            OrderSummary(price="0.3", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.4)

        positions = [
            OrderSummary(price="0.5", size="120"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.3", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.4)

        positions = [
            OrderSummary(price="0.5", size="200"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.3", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.5)

    def test_calculate_market_price_sell(self):
        # empty
        with self.assertRaises(Exception):
            builder = OrderBuilder(signer)
            builder.calculate_market_price([], 100)

        # not enough
        with self.assertRaises(Exception):
            positions = [
                OrderSummary(price="0.4", size="100"),
                OrderSummary(price="0.5", size="100"),
            ]
            builder = OrderBuilder(signer)
            builder.calculate_market_price(positions, 100)

        # OK
        positions = [
            OrderSummary(price="0.3", size="100"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.5", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.5)

        positions = [
            OrderSummary(price="0.3", size="100"),
            OrderSummary(price="0.4", size="300"),
            OrderSummary(price="0.5", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.4)

        positions = [
            OrderSummary(price="0.3", size="100"),
            OrderSummary(price="0.4", size="200"),
            OrderSummary(price="0.5", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.4)

        positions = [
            OrderSummary(price="0.3", size="300"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.5", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.4)

        positions = [
            OrderSummary(price="0.3", size="334"),
            OrderSummary(price="0.4", size="100"),
            OrderSummary(price="0.5", size="100"),
        ]
        builder = OrderBuilder(signer)
        self.assertEqual(builder.calculate_market_price(positions, 100), 0.3)

    def test_get_market_order_amounts_0_1(self):
        builder = OrderBuilder(signer)

        delta_price = 0.1
        delta_size = 0.01
        amount = 0.01
        while amount <= 1000:
            price = 0.1
            while price <= 1:
                maker, taker = builder.get_market_order_amounts(
                    amount, price, ROUNDING_CONFIG["0.1"]
                )
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 2), round_normal(price, 2)
                )
                price = price + delta_price

            amount = amount + delta_size

    def test_get_market_order_amounts_0_01(self):
        builder = OrderBuilder(signer)

        delta_price = 0.01
        delta_size = 0.01
        amount = 0.01
        while amount <= 100:
            price = 0.01
            while price <= 1:
                maker, taker = builder.get_market_order_amounts(
                    amount, price, ROUNDING_CONFIG["0.01"]
                )
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 4), round_normal(price, 4)
                )
                price = price + delta_price

            amount = amount + delta_size

    def test_get_market_order_amounts_0_001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.001
        delta_size = 0.01
        amount = 0.01
        while amount <= 10:
            price = 0.001
            while price <= 1:
                maker, taker = builder.get_market_order_amounts(
                    amount, price, ROUNDING_CONFIG["0.001"]
                )
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 6), round_normal(price, 6)
                )
                price = price + delta_price

            amount = amount + delta_size

    def test_get_market_order_amounts_0_0001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.0001
        delta_size = 0.01
        amount = 0.01
        while amount <= 1:
            price = 0.0001
            while price <= 1:
                maker, taker = builder.get_market_order_amounts(
                    amount, price, ROUNDING_CONFIG["0.0001"]
                )
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 8), round_normal(price, 8)
                )
                price = price + delta_price

            amount = amount + delta_size

    def test_get_order_amounts_buy_0_1(self):
        builder = OrderBuilder(signer)

        delta_price = 0.1
        delta_size = 0.01
        size = 0.01
        while size <= 1000:
            price = 0.1
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    BUY, size, price, ROUNDING_CONFIG["0.1"]
                )
                self.assertEqual(side, 0)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 2), round_normal(price, 2)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_buy_0_01(self):
        builder = OrderBuilder(signer)

        delta_price = 0.01
        delta_size = 0.01
        size = 0.01
        while size <= 100:
            price = 0.01
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    BUY, size, price, ROUNDING_CONFIG["0.01"]
                )
                self.assertEqual(side, 0)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 4), round_normal(price, 4)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_buy_0_001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.001
        delta_size = 0.01
        size = 0.01
        while size <= 10:
            price = 0.001
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    BUY, size, price, ROUNDING_CONFIG["0.001"]
                )
                self.assertEqual(side, 0)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 6), round_normal(price, 6)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_buy_0_0001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.0001
        delta_size = 0.01
        size = 0.01
        while size <= 1:
            price = 0.0001
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    BUY, size, price, ROUNDING_CONFIG["0.0001"]
                )
                self.assertEqual(side, 0)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(maker / taker, 8), round_normal(price, 8)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_sell_0_1(self):
        builder = OrderBuilder(signer)

        delta_price = 0.1
        delta_size = 0.01
        size = 0.01
        while size <= 1000:
            price = 0.1
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    SELL, size, price, ROUNDING_CONFIG["0.1"]
                )
                self.assertEqual(side, 1)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(taker / maker, 2), round_normal(price, 2)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_sell_0_01(self):
        builder = OrderBuilder(signer)

        delta_price = 0.01
        delta_size = 0.01
        size = 0.01
        while size <= 100:
            price = 0.01
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    SELL, size, price, ROUNDING_CONFIG["0.01"]
                )
                self.assertEqual(side, 1)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(taker / maker, 4), round_normal(price, 4)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_sell_0_001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.001
        delta_size = 0.01
        size = 0.01
        while size <= 10:
            price = 0.001
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    SELL, size, price, ROUNDING_CONFIG["0.001"]
                )
                self.assertEqual(side, 1)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(taker / maker, 6), round_normal(price, 6)
                )
                price = price + delta_price

            size = size + delta_size

    def test_get_order_amounts_sell_0_0001(self):
        builder = OrderBuilder(signer)

        delta_price = 0.0001
        delta_size = 0.01
        size = 0.01
        while size <= 1:
            price = 0.0001
            while price <= 1:
                side, maker, taker = builder.get_order_amounts(
                    SELL, size, price, ROUNDING_CONFIG["0.0001"]
                )
                self.assertEqual(side, 1)
                self.assertEqual(decimal_places(maker), 0)
                self.assertEqual(decimal_places(taker), 0)
                self.assertGreaterEqual(
                    round_normal(taker / maker, 8), round_normal(price, 8)
                )
                price = price + delta_price

            size = size + delta_size

    def test_create_order_decimal_accuracy(self):
        builder = OrderBuilder(signer)

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.24,
                size=15,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            3600000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            15000000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.24,
                size=15,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            15000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            3600000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.82,
                size=101,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            82820000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            101000000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.82,
                size=101,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            101000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            82820000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.78,
                size=12.8205,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            9999600,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            12820000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.78,
                size=12.8205,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            12820000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            9999600,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.39,
                size=2435.89,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            2435890000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            949997100,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.43,
                size=19.1,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            19100000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            8213000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.58,
                size=18233.33,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            10575331400,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            18233330000,
        )
        self.assertEqual(
            signed_order.order["makerAmount"] / signed_order.order["takerAmount"], 0.58
        )

    def test_create_order_buy_0_1(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.5,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.1"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            10520000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.5,
        )

    def test_create_order_buy_0_01(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            11782400,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.56,
        )

    def test_create_order_buy_0_001(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.056,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            1178240,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.056,
        )

    def test_create_order_buy_0_0001(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.0056,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.0001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            117824,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.0056,
        )

    def test_create_order_sell_0_1(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.5,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.1"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            10520000,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.5,
        )

    def test_create_order_sell_0_01(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            11782400,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.56,
        )

    def test_create_order_sell_0_001(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.056,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            1178240,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.056,
        )

    def test_create_order_sell_0_0001(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.0056,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.0001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            117824,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.0056,
        )

    def test_create_order_decimal_accuracy_neg_risk(self):
        builder = OrderBuilder(signer)

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.24,
                size=15,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            3600000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            15000000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.24,
                size=15,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            15000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            3600000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.82,
                size=101,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            82820000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            101000000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.82,
                size=101,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            101000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            82820000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.78,
                size=12.8205,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            9999600,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            12820000,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.78,
                size=12.8205,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            12820000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            9999600,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.39,
                size=2435.89,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            2435890000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            949997100,
        )

        # SELL
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.43,
                size=19.1,
                side=SELL,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            19100000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            8213000,
        )

        # BUY
        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.58,
                size=18233.33,
                side=BUY,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            10575331400,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            18233330000,
        )
        self.assertEqual(
            signed_order.order["makerAmount"] / signed_order.order["takerAmount"], 0.58
        )

    def test_create_order_buy_0_1_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.5,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            10520000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.5,
        )

    def test_create_order_buy_0_01_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            11782400,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.56,
        )

    def test_create_order_buy_0_001_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.056,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            1178240,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.056,
        )

    def test_create_order_buy_0_0001_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.0056,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            117824,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.0056,
        )

    def test_create_order_sell_0_1_neg_risk(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.5,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            10520000,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.5,
        )

    def test_create_order_sell_0_01_neg_risk(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            11782400,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.56,
        )

    def test_create_order_sell_0_001_neg_risk(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.056,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            1178240,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.056,
        )

    def test_create_order_sell_0_0001_neg_risk(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.0056,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            21040000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            117824,
        )
        self.assertEqual(
            signed_order.order["side"],
            1,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            50000,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertEqual(
            float(signed_order.order["takerAmount"])
            / float(signed_order.order["makerAmount"]),
            0.0056,
        )

    def test_dict_order_buy_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )

        self.assertIsNotNone(signed_order)

        signed_order_dict = signed_order.dict()

        self.assertIsNotNone(signed_order_dict)
        self.assertTrue(isinstance(signed_order_dict["salt"], int))
        self.assertEqual(
            signed_order_dict["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order_dict["tokenId"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["makerAmount"],
            "11782400",
        )
        self.assertEqual(
            signed_order_dict["takerAmount"],
            "21040000",
        )
        self.assertEqual(
            signed_order_dict["side"],
            BUY,
        )
        self.assertEqual(
            signed_order_dict["expiration"],
            "50000",
        )
        self.assertEqual(
            signed_order_dict["nonce"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["feeRateBps"],
            "111",
        )
        self.assertEqual(
            signed_order_dict["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order_dict["signature"])

    def test_dict_order_sell_neg_risk(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )

        self.assertIsNotNone(signed_order)

        signed_order_dict = signed_order.dict()

        self.assertIsNotNone(signed_order_dict)
        self.assertTrue(isinstance(signed_order_dict["salt"], int))
        self.assertEqual(
            signed_order_dict["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order_dict["tokenId"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["makerAmount"],
            "21040000",
        )
        self.assertEqual(
            signed_order_dict["takerAmount"],
            "11782400",
        )
        self.assertEqual(
            signed_order_dict["side"],
            SELL,
        )
        self.assertEqual(
            signed_order_dict["expiration"],
            "50000",
        )
        self.assertEqual(
            signed_order_dict["nonce"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["feeRateBps"],
            "111",
        )
        self.assertEqual(
            signed_order_dict["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order_dict["signature"])

    def test_dict_order_buy(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=BUY,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )

        self.assertIsNotNone(signed_order)

        signed_order_dict = signed_order.dict()

        self.assertIsNotNone(signed_order_dict)
        self.assertTrue(isinstance(signed_order_dict["salt"], int))
        self.assertEqual(
            signed_order_dict["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order_dict["tokenId"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["makerAmount"],
            "11782400",
        )
        self.assertEqual(
            signed_order_dict["takerAmount"],
            "21040000",
        )
        self.assertEqual(
            signed_order_dict["side"],
            BUY,
        )
        self.assertEqual(
            signed_order_dict["expiration"],
            "50000",
        )
        self.assertEqual(
            signed_order_dict["nonce"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["feeRateBps"],
            "111",
        )
        self.assertEqual(
            signed_order_dict["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order_dict["signature"])

    def test_dict_order_sell(self):
        builder = OrderBuilder(signer, sig_type=POLY_GNOSIS_SAFE)

        signed_order = builder.create_order(
            order_args=OrderArgs(
                token_id="123",
                price=0.56,
                size=21.04,
                side=SELL,
                fee_rate_bps=111,
                nonce=123,
                expiration=50000,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )

        self.assertIsNotNone(signed_order)

        signed_order_dict = signed_order.dict()

        self.assertIsNotNone(signed_order_dict)
        self.assertTrue(isinstance(signed_order_dict["salt"], int))
        self.assertEqual(
            signed_order_dict["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order_dict["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order_dict["tokenId"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["makerAmount"],
            "21040000",
        )
        self.assertEqual(
            signed_order_dict["takerAmount"],
            "11782400",
        )
        self.assertEqual(
            signed_order_dict["side"],
            SELL,
        )
        self.assertEqual(
            signed_order_dict["expiration"],
            "50000",
        )
        self.assertEqual(
            signed_order_dict["nonce"],
            "123",
        )
        self.assertEqual(
            signed_order_dict["feeRateBps"],
            "111",
        )
        self.assertEqual(
            signed_order_dict["signatureType"],
            POLY_GNOSIS_SAFE,
        )
        self.assertIsNotNone(signed_order_dict["signature"])

    def test_create_market_order_buy_0_1(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.5,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.1"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            200000000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.5,
        )

    def test_create_market_order_buy_0_01(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.56,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.01"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            178571400,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.56,
        )

    def test_create_market_order_buy_0_001(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.056,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            1785714280,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.056,
        )

    def test_create_market_order_buy_0_0001(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.0056,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.0001"),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            17857142857,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.0056,
        )

    def test_create_market_order_buy_0_1_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.5,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            200000000,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.5,
        )

    def test_create_market_order_buy_0_01_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.56,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            178571400,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.56,
        )

    def test_create_market_order_buy_0_001_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.056,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            1785714280,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.056,
        )

    def test_create_market_order_buy_0_0001_neg_risk(self):
        builder = OrderBuilder(signer)

        signed_order = builder.create_market_order(
            order_args=MarketOrderArgs(
                token_id="123",
                price=0.0056,
                amount=100,
                fee_rate_bps=111,
                nonce=123,
            ),
            options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
        )

        self.assertTrue(isinstance(signed_order.order["salt"], int))
        self.assertIsNotNone(signed_order)
        self.assertEqual(
            signed_order.order["maker"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["signer"],
            "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
        )
        self.assertEqual(
            signed_order.order["taker"],
            "0x0000000000000000000000000000000000000000",
        )
        self.assertEqual(
            signed_order.order["tokenId"],
            123,
        )
        self.assertEqual(
            signed_order.order["makerAmount"],
            100000000,
        )
        self.assertEqual(
            signed_order.order["takerAmount"],
            17857142857,
        )
        self.assertEqual(
            signed_order.order["side"],
            0,
        )
        self.assertEqual(
            signed_order.order["expiration"],
            0,
        )
        self.assertEqual(
            signed_order.order["nonce"],
            123,
        )
        self.assertEqual(
            signed_order.order["feeRateBps"],
            111,
        )
        self.assertEqual(
            signed_order.order["signatureType"],
            EOA,
        )
        self.assertIsNotNone(signed_order.signature)
        self.assertGreaterEqual(
            float(signed_order.order["makerAmount"])
            / float(signed_order.order["takerAmount"]),
            0.0056,
        )



from unittest import TestCase

from py_clob_client.order_builder.helpers import decimal_places


class TestHelpers(TestCase):
    def test_decimal_places(self):
        self.assertEqual(decimal_places(949.9970999999999), 13)
        self.assertEqual(decimal_places(949), 0)



import requests

from py_clob_client.clob_types import (
    DropNotificationParams,
    BalanceAllowanceParams,
    OrderScoringParams,
    OrdersScoringParams,
    TradeParams,
    OpenOrderParams,
)

from ..exceptions import PolyApiException

GET = "GET"
POST = "POST"
DELETE = "DELETE"
PUT = "PUT"


def overloadHeaders(method: str, headers: dict) -> dict:
    if headers is None:
        headers = dict()
    headers["User-Agent"] = "py_clob_client"

    headers["Accept"] = "*/*"
    headers["Connection"] = "keep-alive"
    headers["Content-Type"] = "application/json"

    if method == GET:
        headers["Accept-Encoding"] = "gzip"

    return headers


def request(endpoint: str, method: str, headers=None, data=None):
    try:
        headers = overloadHeaders(method, headers)
        resp = requests.request(
            method=method, url=endpoint, headers=headers, json=data if data else None
        )
        if resp.status_code != 200:
            raise PolyApiException(resp)

        try:
            return resp.json()
        except requests.JSONDecodeError:
            return resp.text

    except requests.RequestException:
        raise PolyApiException(error_msg="Request exception!")


def post(endpoint, headers=None, data=None):
    return request(endpoint, POST, headers, data)


def get(endpoint, headers=None, data=None):
    return request(endpoint, GET, headers, data)


def delete(endpoint, headers=None, data=None):
    return request(endpoint, DELETE, headers, data)


def build_query_params(url: str, param: str, val: str) -> str:
    url_with_params = url
    last = url_with_params[-1]
    # if last character in url string == "?", append the param directly: api.com?param=value
    if last == "?":
        url_with_params = "{}{}={}".format(url_with_params, param, val)
    else:
        # else add "&", then append the param
        url_with_params = "{}&{}={}".format(url_with_params, param, val)
    return url_with_params


def add_query_trade_params(
    base_url: str, params: TradeParams = None, next_cursor="MA=="
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.market:
            url = build_query_params(url, "market", params.market)
        if params.asset_id:
            url = build_query_params(url, "asset_id", params.asset_id)
        if params.after:
            url = build_query_params(url, "after", params.after)
        if params.before:
            url = build_query_params(url, "before", params.before)
        if params.maker_address:
            url = build_query_params(url, "maker_address", params.maker_address)
        if params.id:
            url = build_query_params(url, "id", params.id)
        if next_cursor:
            url = build_query_params(url, "next_cursor", next_cursor)
    return url


def add_query_open_orders_params(
    base_url: str, params: OpenOrderParams = None, next_cursor="MA=="
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.market:
            url = build_query_params(url, "market", params.market)
        if params.asset_id:
            url = build_query_params(url, "asset_id", params.asset_id)
        if params.id:
            url = build_query_params(url, "id", params.id)
        if next_cursor:
            url = build_query_params(url, "next_cursor", next_cursor)
    return url


def drop_notifications_query_params(
    base_url: str, params: DropNotificationParams = None
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.ids:
            url = build_query_params(url, "ids", ",".join(params.ids))
    return url


def add_balance_allowance_params_to_url(
    base_url: str, params: BalanceAllowanceParams = None
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.asset_type:
            url = build_query_params(url, "asset_type", params.asset_type.__str__())
        if params.token_id:
            url = build_query_params(url, "token_id", params.token_id)
        if params.signature_type is not None:
            url = build_query_params(url, "signature_type", params.signature_type)
    return url


def add_order_scoring_params_to_url(
    base_url: str, params: OrderScoringParams = None
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.orderId:
            url = build_query_params(url, "order_id", params.orderId)
    return url


def add_orders_scoring_params_to_url(
    base_url: str, params: OrdersScoringParams = None
) -> str:
    """
    Adds query parameters to a url
    """
    url = base_url
    if params:
        url = url + "?"
        if params.orderIds:
            url = build_query_params(url, "order_ids", ",".join(params.orderIds))
    return url



from dataclasses import dataclass
from typing import Any
from dataclasses import dataclass, asdict
from json import dumps
from typing import Literal, Optional

from .constants import ZERO_ADDRESS


@dataclass
class ApiCreds:
    api_key: str
    api_secret: str
    api_passphrase: str


@dataclass
class RequestArgs:
    method: str
    request_path: str
    body: Any = None


@dataclass
class BookParams:
    token_id: str
    side: str = ""


@dataclass
class OrderArgs:
    token_id: str
    """
    TokenID of the Conditional token asset being traded
    """

    price: float
    """
    Price used to create the order
    """

    size: float
    """
    Size in terms of the ConditionalToken
    """

    side: str
    """
    Side of the order
    """

    fee_rate_bps: int = 0
    """
    Fee rate, in basis points, charged to the order maker, charged on proceeds
    """

    nonce: int = 0
    """
    Nonce used for onchain cancellations
    """

    expiration: int = 0
    """
    Timestamp after which the order is expired.
    """

    taker: str = ZERO_ADDRESS
    """
    Address of the order taker. The zero address is used to indicate a public order
    """


@dataclass
class MarketOrderArgs:
    token_id: str
    """
    TokenID of the Conditional token asset being traded
    """

    amount: float
    """
    Amount in terms of Collateral
    """

    price: float = 0
    """
    Price used to create the order
    """

    fee_rate_bps: int = 0
    """
    Fee rate, in basis points, charged to the order maker, charged on proceeds
    """

    nonce: int = 0
    """
    Nonce used for onchain cancellations
    """

    taker: str = ZERO_ADDRESS
    """
    Address of the order taker. The zero address is used to indicate a public order
    """


@dataclass
class TradeParams:
    id: str = None
    maker_address: str = None
    market: str = None
    asset_id: str = None
    before: int = None
    after: int = None


@dataclass
class OpenOrderParams:
    id: str = None
    market: str = None
    asset_id: str = None


@dataclass
class DropNotificationParams:
    ids: list[str] = None


@dataclass
class OrderSummary:
    price: str = None
    size: str = None

    @property
    def __dict__(self):
        return asdict(self)

    @property
    def json(self):
        return dumps(self.__dict__)


@dataclass
class OrderBookSummary:
    market: str = None
    asset_id: str = None
    bids: list[OrderSummary] = None
    asks: list[OrderSummary] = None
    hash: str = None

    @property
    def __dict__(self):
        return asdict(self)

    @property
    def json(self):
        return dumps(self.__dict__, separators=(",", ":"))


class AssetType(enumerate):
    COLLATERAL = "COLLATERAL"
    CONDITIONAL = "CONDITIONAL"


@dataclass
class BalanceAllowanceParams:
    asset_type: AssetType = None
    token_id: str = None
    signature_type: int = -1


class OrderType(enumerate):
    GTC = "GTC"
    FOK = "FOK"
    GTD = "GTD"


@dataclass
class OrderScoringParams:
    orderId: str


@dataclass
class OrdersScoringParams:
    orderIds: list[str]


TickSize = Literal["0.1", "0.01", "0.001", "0.0001"]


@dataclass
class CreateOrderOptions:
    tick_size: TickSize
    neg_risk: bool = False


@dataclass
class PartialCreateOrderOptions:
    tick_size: Optional[TickSize] = None
    neg_risk: bool = False


@dataclass
class RoundConfig:
    price: float
    size: float
    amount: float


@dataclass
class ContractConfig:
    """
    Contract Configuration
    """

    exchange: str
    """
    The exchange contract responsible for matching orders
    """

    collateral: str
    """
    The ERC20 token used as collateral for the exchange's markets
    """

    conditional_tokens: str
    """
    The ERC1155 conditional tokens contract
    """



from unittest import TestCase

from py_clob_client.clob_types import (
    OrderArgs,
    OrderType,
    CreateOrderOptions,
)
from py_clob_client.constants import AMOY
from py_clob_client.order_builder.constants import BUY, SELL
from py_clob_client.signer import Signer
from py_clob_client.order_builder.builder import OrderBuilder
from py_clob_client.utilities import (
    parse_raw_orderbook_summary,
    generate_orderbook_summary_hash,
    order_to_json,
    is_tick_size_smaller,
    price_valid,
)


class TestUtilities(TestCase):
    def test_parse_raw_orderbook_summary(self):
        raw_obs = {
            "market": "0xbd31dc8a20211944f6b70f31557f1001557b59905b7738480ca09bd4532f84af",
            "asset_id": "52114319501245915516055106046884209969926127482827954674443846427813813222426",
            "bids": [
                {"price": "0.15", "size": "100"},
                {"price": "0.31", "size": "148.56"},
                {"price": "0.33", "size": "58"},
                {"price": "0.5", "size": "100"},
            ],
            "asks": [],
            "hash": "9d6d9e8831a150ac4cd878f99f7b2c6d419b875f",
        }

        orderbook_summary = parse_raw_orderbook_summary(raw_obs)

        self.assertEqual(
            orderbook_summary.market,
            "0xbd31dc8a20211944f6b70f31557f1001557b59905b7738480ca09bd4532f84af",
        )
        self.assertEqual(
            orderbook_summary.asset_id,
            "52114319501245915516055106046884209969926127482827954674443846427813813222426",
        )
        self.assertEqual(
            orderbook_summary.hash, "9d6d9e8831a150ac4cd878f99f7b2c6d419b875f"
        )

        self.assertIsNotNone(orderbook_summary.asks)
        self.assertIsNotNone(orderbook_summary.bids)

        self.assertEqual(len(orderbook_summary.asks), 0)
        self.assertEqual(len(orderbook_summary.bids), 4)

        self.assertEqual(orderbook_summary.bids[0].price, "0.15")
        self.assertEqual(orderbook_summary.bids[0].size, "100")
        self.assertEqual(orderbook_summary.bids[1].price, "0.31")
        self.assertEqual(orderbook_summary.bids[1].size, "148.56")
        self.assertEqual(orderbook_summary.bids[2].price, "0.33")
        self.assertEqual(orderbook_summary.bids[2].size, "58")
        self.assertEqual(orderbook_summary.bids[3].price, "0.5")
        self.assertEqual(orderbook_summary.bids[3].size, "100")

        raw_obs = {
            "market": "0xaabbcc",
            "asset_id": "100",
            "bids": [],
            "asks": [],
            "hash": "7f81a35a09e1933a96b05edb51ac4be4a6163146",
        }

        orderbook_summary = parse_raw_orderbook_summary(raw_obs)

        self.assertEqual(
            orderbook_summary.market,
            "0xaabbcc",
        )
        self.assertEqual(
            orderbook_summary.asset_id,
            "100",
        )
        self.assertEqual(
            orderbook_summary.hash, "7f81a35a09e1933a96b05edb51ac4be4a6163146"
        )

        self.assertIsNotNone(orderbook_summary.asks)
        self.assertIsNotNone(orderbook_summary.bids)

        self.assertEqual(len(orderbook_summary.asks), 0)
        self.assertEqual(len(orderbook_summary.bids), 0)

    def test_generate_orderbook_summary_hash(self):
        raw_obs = {
            "market": "0xaabbcc",
            "asset_id": "100",
            "bids": [
                {"price": "0.3", "size": "100"},
                {"price": "0.4", "size": "100"},
            ],
            "asks": [
                {"price": "0.6", "size": "100"},
                {"price": "0.7", "size": "100"},
            ],
            "hash": "",
        }

        orderbook_summary = parse_raw_orderbook_summary(raw_obs)
        self.assertEqual(
            generate_orderbook_summary_hash(orderbook_summary),
            "b8b72c72c6534d1b3a4e7fb47b81672d0e94d5a5",
        )
        self.assertEqual(
            orderbook_summary.hash,
            "b8b72c72c6534d1b3a4e7fb47b81672d0e94d5a5",
        )

        raw_obs = {
            "market": "0xaabbcc",
            "asset_id": "100",
            "bids": [
                {"price": "0.3", "size": "100"},
                {"price": "0.4", "size": "100"},
            ],
            "asks": [
                {"price": "0.6", "size": "100"},
                {"price": "0.7", "size": "100"},
            ],
            "hash": "b8b72c72c6534d1b3a4e7fb47b81672d0e94d5a5",
        }

        orderbook_summary = parse_raw_orderbook_summary(raw_obs)
        self.assertEqual(
            generate_orderbook_summary_hash(orderbook_summary),
            "b8b72c72c6534d1b3a4e7fb47b81672d0e94d5a5",
        )
        self.assertEqual(
            orderbook_summary.hash,
            "b8b72c72c6534d1b3a4e7fb47b81672d0e94d5a5",
        )

        raw_obs = {
            "market": "0xaabbcc",
            "asset_id": "100",
            "bids": [],
            "asks": [],
            "hash": "",
        }

        orderbook_summary = parse_raw_orderbook_summary(raw_obs)
        self.assertEqual(
            generate_orderbook_summary_hash(orderbook_summary),
            "7f81a35a09e1933a96b05edb51ac4be4a6163146",
        )
        self.assertEqual(
            orderbook_summary.hash,
            "7f81a35a09e1933a96b05edb51ac4be4a6163146",
        )

    def test_order_to_json_0_1(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.1"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.1"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.1"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.1"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_01(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.01"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "5000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.01"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "5000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.01"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "5000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.01"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "5000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_001(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.001"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "500000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.001"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "500000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.001"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "500000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.001"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "500000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_0001(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.0001"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.0001"),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.0001"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.0001"),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_1_neg_risk(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.5,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.1", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_01_neg_risk(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "5000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "5000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "5000000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.05,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.01", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "5000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_001_neg_risk(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "500000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "500000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "500000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "500000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_order_to_json_0_0001_neg_risk(self):
        # publicly known private key
        private_key = (
            "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
        )
        chain_id = AMOY
        signer = Signer(private_key=private_key, chain_id=chain_id)
        owner = "aaa-bbb-ccc"
        builder = OrderBuilder(signer)

        # GTC BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTC SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTC,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTC")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD BUY
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=BUY,
                ),
                options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "50000")
        self.assertEqual(json_order["order"]["takerAmount"], "100000000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "BUY")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

        # GTD SELL
        json_order = order_to_json(
            order=builder.create_order(
                order_args=OrderArgs(
                    token_id="100",
                    price=0.0005,
                    size=100,
                    side=SELL,
                ),
                options=CreateOrderOptions(tick_size="0.0001", neg_risk=True),
            ),
            owner=owner,
            orderType=OrderType.GTD,
        )

        self.assertIsNotNone(json_order)
        self.assertEqual(json_order["orderType"], "GTD")
        self.assertEqual(json_order["owner"], owner)
        self.assertIsNotNone(json_order["order"])
        self.assertIsNotNone(json_order["order"]["salt"])
        self.assertEqual(
            json_order["order"]["maker"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["signer"], "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
        )
        self.assertEqual(
            json_order["order"]["taker"], "0x0000000000000000000000000000000000000000"
        )
        self.assertEqual(json_order["order"]["tokenId"], "100")
        self.assertEqual(json_order["order"]["makerAmount"], "100000000")
        self.assertEqual(json_order["order"]["takerAmount"], "50000")
        self.assertEqual(json_order["order"]["expiration"], "0")
        self.assertEqual(json_order["order"]["nonce"], "0")
        self.assertEqual(json_order["order"]["feeRateBps"], "0")
        self.assertEqual(json_order["order"]["side"], "SELL")
        self.assertEqual(json_order["order"]["signatureType"], 0)
        self.assertIsNotNone(json_order["order"]["signature"])

    def test_is_tick_size_smaller(self):
        # 0.1
        self.assertFalse(is_tick_size_smaller("0.1", "0.1"))
        self.assertFalse(is_tick_size_smaller("0.1", "0.01"))
        self.assertFalse(is_tick_size_smaller("0.1", "0.001"))
        self.assertFalse(is_tick_size_smaller("0.1", "0.0001"))

        # 0.01
        self.assertTrue(is_tick_size_smaller("0.01", "0.1"))
        self.assertFalse(is_tick_size_smaller("0.01", "0.01"))
        self.assertFalse(is_tick_size_smaller("0.01", "0.001"))
        self.assertFalse(is_tick_size_smaller("0.01", "0.0001"))

        # 0.001
        self.assertTrue(is_tick_size_smaller("0.001", "0.1"))
        self.assertTrue(is_tick_size_smaller("0.001", "0.01"))
        self.assertFalse(is_tick_size_smaller("0.001", "0.001"))
        self.assertFalse(is_tick_size_smaller("0.001", "0.0001"))

        # 0.0001
        self.assertTrue(is_tick_size_smaller("0.0001", "0.1"))
        self.assertTrue(is_tick_size_smaller("0.0001", "0.01"))
        self.assertTrue(is_tick_size_smaller("0.0001", "0.001"))
        self.assertFalse(is_tick_size_smaller("0.0001", "0.0001"))

    def test_price_valid(self):
        self.assertFalse(price_valid(0.00001, "0.0001"))
        self.assertTrue(price_valid(0.0001, "0.0001"))
        self.assertTrue(price_valid(0.001, "0.0001"))
        self.assertTrue(price_valid(0.01, "0.0001"))
        self.assertTrue(price_valid(0.1, "0.0001"))
        self.assertTrue(price_valid(0.9, "0.0001"))
        self.assertTrue(price_valid(0.99, "0.0001"))
        self.assertTrue(price_valid(0.999, "0.0001"))
        self.assertTrue(price_valid(0.9999, "0.0001"))
        self.assertFalse(price_valid(0.99999, "0.0001"))

        self.assertFalse(price_valid(0.00001, "0.001"))
        self.assertFalse(price_valid(0.0001, "0.001"))
        self.assertTrue(price_valid(0.001, "0.001"))
        self.assertTrue(price_valid(0.01, "0.001"))
        self.assertTrue(price_valid(0.1, "0.001"))
        self.assertTrue(price_valid(0.9, "0.001"))
        self.assertTrue(price_valid(0.99, "0.001"))
        self.assertTrue(price_valid(0.999, "0.001"))
        self.assertFalse(price_valid(0.9999, "0.001"))
        self.assertFalse(price_valid(0.99999, "0.001"))

        self.assertFalse(price_valid(0.00001, "0.01"))
        self.assertFalse(price_valid(0.0001, "0.01"))
        self.assertFalse(price_valid(0.001, "0.01"))
        self.assertTrue(price_valid(0.01, "0.01"))
        self.assertTrue(price_valid(0.1, "0.01"))
        self.assertTrue(price_valid(0.9, "0.01"))
        self.assertTrue(price_valid(0.99, "0.01"))
        self.assertFalse(price_valid(0.999, "0.01"))
        self.assertFalse(price_valid(0.9999, "0.01"))
        self.assertFalse(price_valid(0.99999, "0.01"))

        self.assertFalse(price_valid(0.00001, "0.1"))
        self.assertFalse(price_valid(0.0001, "0.1"))
        self.assertFalse(price_valid(0.001, "0.1"))
        self.assertFalse(price_valid(0.01, "0.1"))
        self.assertTrue(price_valid(0.1, "0.1"))
        self.assertTrue(price_valid(0.9, "0.1"))
        self.assertFalse(price_valid(0.99, "0.1"))
        self.assertFalse(price_valid(0.999, "0.1"))
        self.assertFalse(price_valid(0.9999, "0.1"))
        self.assertFalse(price_valid(0.99999, "0.1"))
